"""
This package provides a checker for NuSMV models and CTL properties without
fairness. This means that, even the model provides some fairness constraints,
CTL is checked without taking them into account.
"""